package org.anarres.jperf;

import java.util.AbstractMap;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;

public class PerfectStringMapGenerator<V>
					extends PerfectMapGenerator<String,V> {

	private static final class Meta<V> {
		char[]	ksig;
		int		klen;
		int		kfreq;
		V		value;

		public Meta(char[] ksig, int klen, V value) {
			this.ksig = ksig;
			this.klen = klen;
			this.kfreq = 0;
			this.value = value;
		}
		public String toString() {
			StringBuilder	buf = new StringBuilder();
			buf.append(Arrays.toString(ksig));
			buf.append(", klen=").append(klen);
			buf.append(", kfreq=").append(kfreq);
			return buf.toString();
		}
	}

	/** Shortens the array to the given length. */
	private static final char[] shorten(char[] arr, int len) {
		char[]	tmp = new char[len];
		System.arraycopy(arr, 0, tmp, 0, len);
		return tmp;
	}

	/** Shortens the array to the given length. */
	private static final int[] shorten(int[] arr, int len) {
		int[]	tmp = new int[len];
		System.arraycopy(arr, 0, tmp, 0, len);
		return tmp;
	}

	/** Sorts arr according to freq. */
	private static final void sort(char[] arr, int[] freq) {
		/* Insertion sort is fastest for small arrays. */
		for (int i = 1; i < arr.length; i++) {	// XXX JVM does 0!
			for (int j = i; j > 0; j--) {
				if (freq[arr[j]] < freq[arr[j - 1]]) {
					char	tmp = arr[j];
					arr[j] = arr[j - 1];
					arr[j - 1] = tmp;
				}
				else {
					break;
				}
			}
		}
	}

/* Instance. */

	private Map<String,Meta<V>>	map;
	private byte[]				keys;

	public PerfectStringMapGenerator(byte[] keys) {
		this.map = new HashMap<String,Meta<V>>();
		Arrays.sort(keys);
		this.keys = keys;
	}

	public void add(String k, V v) {
		char[]	ksig = new char[keys.length];
		int		klen = k.length();
		int		idx = 0;
		for (byte b : keys) {
			if (b >= 0) {
				if (b >= klen)
					continue;
				ksig[idx++] = k.charAt(b);
			}
			else {
				int	kidx = klen + b;	// b is negative
				if (kidx < 0)
					continue;
				if (kidx >= klen)
					continue;
				ksig[idx++] = k.charAt(kidx);
			}
		}
		if (idx < ksig.length)
			ksig = shorten(ksig, idx);
		Meta<V>		prev = map.put(k, new Meta(ksig, klen, v));
		if (prev != null)
			throw new IllegalArgumentException("Duplicate key " + k +
					", previous value was " + prev.value +
					", new value is " + v);
	}

	/** This method sucks less than it used to. */
	private int[] buildFrequencyTable() {
		/* This array is size 64K * sizeof(int) */
		// XXX int[]	freq = new int[Character.MAX_VALUE];
		int[]	freq = new int[Character.MAX_VALUE];
		Arrays.fill(freq, 0);
		for (Meta<V> m : map.values())
			for (char c : m.ksig)
				freq[c]++;
		return freq;
	}

	/** Select and sort Meta by freq. */
	private Meta[] buildMetaOrder(int[] freq) {
		Meta[]	meta = new Meta[map.size()];
		int		idx = 0;
		for (Meta<V> m : map.values()) {
			meta[idx++] = m;
			for (char c : m.ksig) {
				m.kfreq += freq[c];
			}
			sort(m.ksig, freq);
		}
		Arrays.sort(meta, new Comparator<Meta>() {
			public int compare(Meta a, Meta b) {
				return Integer.signum(b.kfreq - a.kfreq);
			}
		});
		System.out.println("Meta order is " + Arrays.toString(meta));
		return meta;
	}

	/** Find a char in a which is not in b. */
	private char diff(char[] a, char[] b) {
		/* Since a and b are sorted ksigs, we expect this to
		 * terminate fast because the least frequent members of a are
		 * earlier. */
		A: for (char _a : a) {
			for (char _b : b)
				if (_a == _b)
					continue A;
			return _a;
		}
		throw new IllegalArgumentException(
			"Failed to find disjoint union of keysigs: " +
			Arrays.toString(a) + " and " + Arrays.toString(b)
				);
	}

	private int hash(Meta<V> m, int[] kvals, char min) {
		int	out = m.klen;
		for (char c : m.ksig)
			out += kvals[c - min];
		return out;
	}

	private void resolve(Meta<V> x, Meta<V> y, int[] kvals, char min) {
		char	c = diff(x.ksig, y.ksig);
		kvals[c - min]++;
	}

	public Map<String,V> toMap() {
		if (map.isEmpty())
			return Collections.<String,V>emptyMap();

		/* XXX Select appropriate/optimal keysigs. */
		int[]		freq = buildFrequencyTable();
		Meta[]		meta = buildMetaOrder(freq);

		char		min = Character.MAX_VALUE;
		char		max = 0;

		for (char i = 0; i < freq.length; i++) {
			if (freq[i] > 0) {
				min = i;
				break;
			}
		}

		for (char i = (char)(freq.length - 1); i >= 0; i--) {
			if (freq[i] > 0) {
				max = i;
				break;
			}
		}

		{
			freq = shorten(freq, max);
			System.out.println("Freq map is " + Arrays.toString(freq));
		}

		if (max < min)
			throw new IllegalStateException(
				"Computation failed. Empty key set or all empty keys?"
					);

		int[]		kvals = new int[max - min + 1];
		Arrays.fill(kvals, 0);

		int			vsize = meta.length;
		/* Add the fiddle factor to the table size. */
		int			tsize = vsize + (vsize >> 1);
		System.out.println("Table size is " + tsize);
		Meta[]		used = new Meta[tsize];

		FUNC: for (;;) {
			Arrays.fill(used, null);
			for (Meta<V> m : meta) {
				int	hash = hash(m, kvals, min);
				System.out.println(m + " => " + hash);
				if (hash >= used.length) {
					tsize = hash + 1;
					System.out.println("Resizing table to " + tsize);
					used = new Meta[tsize];
					continue FUNC;
				}
				if (used[hash] != null) {
					resolve(m, used[hash], kvals, min);
					continue FUNC;
				}
				used[hash] = m;
			}

			Map.Entry[]	table = new Map.Entry[tsize];
			for (Map.Entry<String,Meta<V>> e : map.entrySet()) {
				Meta<V>	m = e.getValue();
				int	hash = hash(m, kvals, min);
				assert table[hash] == null : "shit! fail!";
				table[hash] = new AbstractMap.SimpleEntry(
					e.getKey(),
					m.value
				);
			}
			return new PerfectStringMap(table, kvals, min, keys);
		}
	}

}
